@@ -1,11 +1,12 @@
 /*
- * Extend a tcl command interpreter with commands to query the
- * Internet domain name service.
+ * Extend a tcl command interpreter with a dns command to query 
+ * the Internet domain name service.
  *
  * Copyright (c) 1993, 1994
- *                    E. Schoenfelder, J. Schoenwaelder
- *                    TU Braunschweig, Germany
- *                    Institute for Operating Systems and Computer Networks
+ *
+ * E. Schoenfelder, J. Schoenwaelder
+ * TU Braunschweig, Germany
+ * Institute for Operating Systems and Computer Networks
  *
  * Permission to use, copy, modify, and distribute this
  * software and its documentation for any purpose and without
@@ -48,15 +49,6 @@
  */
 
 
-/* comment next line to get a standalone module: */
-/* #define STANDALONE		/ * define for standalone version */
-
-#if defined(STANDALONE) && ! defined(DNS)
-#define HAVE_DNS
-#endif
-
-#ifdef HAVE_DNS
-
 #include <stdio.h>
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
@@ -75,9 +67,12 @@
 #include <arpa/inet.h>
 #include <resolv.h>
 #include <netdb.h>
-#ifndef STANDALONE
+
 #include <tcl.h>
-#endif
+
+#include "scotty.h"
+#include "xmalloc.h"
+
 
 /* debug switch: */
 static int do_debug = 0;
@@ -101,56 +96,86 @@ static struct in_addr default_server;
  * query and reply structure:
  */
 typedef struct {
-	HEADER qb1;
-	char qb2[PACKETSZ];
+    HEADER qb1;
+    char qb2[PACKETSZ];
 } querybuf;
 
 /*
  * selfmade reply structure (private use only):
  */
 typedef struct {
-	int type;		/* T_A, T_SOA, T_HINFO, T_MX */
-	int n;			/* # of results stored */
-	union {
-	  	struct in_addr addr [MAXRESULT]; 
-		char str [MAXRESULT][256];
-	} u;
+    int type;		/* T_A, T_SOA, T_HINFO, T_MX */
+    int n;			/* # of results stored */
+    union {
+	struct in_addr addr [MAXRESULT]; 
+	char str [MAXRESULT][256];
+    } u;
 } a_res;
 
-
-
 /*
- * failsafe malloc:
+ * Forward declarations for procedures defined later in this file:
  */
 
-static void *
-xmalloc (n)
-int n;
+static int
+is_num_addr		_ANSI_ARGS_((char *s));
+
+static void
+init_stuff		_ANSI_ARGS_((void));
+
+static int
+convert_to_ptr		_ANSI_ARGS_((char *hname, char **res));
+
+static void
+do_query		_ANSI_ARGS_((char *query_string, int query_type, 
+				     a_res *query_result));
+static void
+have_a_query		_ANSI_ARGS_((char *query_string, int query_type,
+				     a_res *query_result, int depth));
+static int 
+dns_a			_ANSI_ARGS_((char *hname, char **ip));
+
+static int
+dns_ptr			_ANSI_ARGS_((char *ip, char **hname));
+
+static void
+hinfo_clean		_ANSI_ARGS_((char *str));
+
+static int
+dns_hinfo		_ANSI_ARGS_((char *hname, char **hinfo));
+
+static int 
+dns_mx			_ANSI_ARGS_((char *hname, char **mx));
+
+static int 
+dns_soa			_ANSI_ARGS_((char *hname, char **soa));
+
+static int
+dns_query		_ANSI_ARGS_((int argc, char **argv, char *buf));
+
+/*
+ * check about numerical address (dotted format).
+ * (i'll assume a ``d.c.b.a.in-addr.arpa'' is not a numerical address.)
+ *
+ * hack alert: the check assumes a name does not have a digit at the 
+ * start and end. 
+ */
+
+static int
+is_num_addr (s)
+     char *s;
 {
-	char *nnew;
+    char *p = s + strlen (s) - 1;
 
-	for (;;) 
-	{
-		nnew = malloc ((unsigned) n);
-		if (nnew)
-			break;
-		sleep (1);
-	}
-	return nnew;
+    if (! isdigit (s [0]))
+	return 0;
+    while (p > s && isspace (*p))
+	p--;
+    if (isdigit (*p))
+	return 1;
+
+    return 0;
 }
 
-static char *
-xstrdup (s)
-char *s;
-{
-	char *nnew = xmalloc (strlen (s) + 1);
-	strcpy (nnew, s);
-	return nnew;
-}
-
-#define xfree(s)	free ((void *) s)
-
-
 
 /*
  * initialize the resolver; set default retries/timeout
@@ -159,19 +184,50 @@ char *s;
 static void
 init_stuff ()
 {
-  int i;
+    int i;
 
-  initialized = 1;
-  res_init ();
-  _res.options |= RES_RECURSE | RES_DNSRCH | RES_DEFNAMES | RES_AAONLY;
+    initialized = 1;
+    res_init ();
+    _res.options |= RES_RECURSE | RES_DNSRCH | RES_DEFNAMES | RES_AAONLY;
 
-  if (_res.nscount > 0)
-    default_server = _res.nsaddr.sin_addr;
+    if (_res.nscount > 0) {
+        default_server = _res.nsaddr.sin_addr;
+    }
   
-  for (i = 0; i < MAXDNSRCH + 1 && _res.dnsrch [i]; i++)
-    dprintf ("* search-order:  %d: `%s'\n", i, _res.dnsrch [i]);
+    for (i = 0; i < MAXDNSRCH + 1 && _res.dnsrch [i]; i++)
+      dprintf ("* search-order:  %d: `%s'\n", i, _res.dnsrch [i]);
 
-  _res.retrans = timeout, _res.retry = retries;
+    _res.retrans = timeout, _res.retry = retries;
+}
+
+
+/*
+ * make ptr from hname (assumed to be numerical).
+ * return -1 on error. 
+ * return in res, malloc'ed string (result or error).
+ */
+
+static int
+convert_to_ptr (hname, res)
+     char *hname, **res;
+{
+    char *result [MAXRESULT];
+    int rc;
+    
+    dprintf ("** convert_to_ptr: first checking ptr...\n");
+    
+    rc = dns_ptr (hname, result);
+    if (rc < 0)
+    {
+	dprintf ("dns_hinfo: cannot reverselookup ``%s''.\n", hname);
+	*res = result [0];
+	return -1;
+    }
+    
+    dprintf ("dns_hinfo: got reverselookup ``%s'' -> ``%s''.\n", hname,
+	     result [0]);
+    *res = result [0];
+    return 0;
 }
 
 
@@ -181,9 +237,9 @@ init_stuff ()
 
 static void
 do_query (query_string, query_type, query_result)
-char *query_string;
-int query_type;
-a_res *query_result;
+     char *query_string;
+     int query_type;
+     a_res *query_result;
 {
 	querybuf query, answer;
 	char buf [512], lbuf [512], auth_buf [512];
@@ -191,7 +247,8 @@ a_res *query_result;
 	short type, class, rdlen;
 	long ttl;
 	querybuf *q;
-	char *ptr, *eom;
+	u_char *ptr;
+	char *eom;
 
 	dprintf ("*** do_query() called:  query_string: `%s'  type: %d\n",
 		query_string, query_type);
@@ -204,21 +261,17 @@ a_res *query_result;
 	 */
 	err = 0;
 
-#if 0
-	bzero (&query, sizeof (querybuf));
-	bzero (&answer, sizeof (answer));
-#else
 	for (i = 0; i < sizeof (querybuf); i++)
 		((char *) &query) [i] = ((char *) &answer) [i] = 0;
-#endif
 
  /**
   ** res_mkquery(op, dname, class, type, data, datalen, newrr, buf, buflen) 
   **/
 	
-	qlen = res_mkquery (QUERY, query_string, C_IN, query_type, (char *) 0, 0, 0,
+	qlen = res_mkquery (QUERY, query_string, C_IN, query_type, 
+			    (char *) 0, 0, 0,
 			    (char *) &query, sizeof(query));
-dprintf ("** res_mkquery() = %d\n", qlen);
+	dprintf ("** res_mkquery() = %d\n", qlen);
 
 	if (qlen <= 0)
 	{
@@ -231,9 +284,10 @@ dprintf ("** res_mkquery() = %d\n", qlen);
   ** res_send(msg, msglen, answer, anslen)
   **/	
 
-	alen = res_send ((char *) &query, qlen, (char *) &answer, sizeof (answer));
+	alen = res_send ((char *) &query, qlen, 
+			 (char *) &answer, sizeof (answer));
 
-dprintf ("** res_send() = %d\n", alen);
+	dprintf ("** res_send() = %d\n", alen);
 
 	if (alen <= 0)
 	{
@@ -275,11 +329,6 @@ dprintf ("** got rcode: %d   is response: %d  opcode: %d\n", q->qb1.rcode,
 		query_result->type = query_type;
 		query_result->n = -1;
 		return;
-#if 0
-		/* dont return; store result only. */
-		dprintf ("* failure - return...\n");
-		return;
-#endif	
 	}
 
 	nscount = ntohs (q->qb1.ancount);
@@ -292,7 +341,7 @@ dprintf ("** got rcode: %d   is response: %d  opcode: %d\n", q->qb1.rcode,
 	eom = (char *) &answer + alen;
 	*eom++ = 0;
 	
-	ptr = q->qb2;
+	ptr = (u_char *) q->qb2;
 
 	/*
 	 * skip over question section: 
@@ -340,11 +389,11 @@ dprintf ("** got rcode: %d   is response: %d  opcode: %d\n", q->qb1.rcode,
 		ptr += llen;
 
 		/* fetch TYPE, CLASS, TTL: */
-		GETSHORT (type, (u_char *) ptr);
-		GETSHORT (class, (u_char *) ptr);
-		GETLONG (ttl, (u_char *) ptr);
+		GETSHORT (type, ptr);
+		GETSHORT (class, ptr);
+		GETLONG (ttl, ptr);
 		/* fetch RDLENGTH: */
-		GETSHORT (rdlen, (u_char *) ptr);
+		GETSHORT (rdlen, ptr);
 
 		dprintf ("** type: %d  class: %d  ttl: %ld  rdlen: %d\n", 
 			 type, class, ttl, (int) rdlen);
@@ -367,7 +416,7 @@ dprintf ("** got rcode: %d   is response: %d  opcode: %d\n", q->qb1.rcode,
 			unsigned long x;
 
 			dprintf ("** type T_A\n");
-			GETLONG (x, (u_char *) ptr);
+			GETLONG (x, ptr);
 			dprintf ("** addr: 0x%08lx\n", x);
 			
 			/* save this address: */
@@ -468,7 +517,7 @@ dprintf ("** got rcode: %d   is response: %d  opcode: %d\n", q->qb1.rcode,
 		else if (type == T_MX)
 		{
 			unsigned prio;
-			GETSHORT (prio, (u_char *) ptr);
+			GETSHORT (prio, ptr);
 
 			len = dn_expand ((char *) q, eom, ptr, buf,
 					 sizeof (buf));
@@ -507,197 +556,192 @@ dprintf ("** got rcode: %d   is response: %d  opcode: %d\n", q->qb1.rcode,
 
 static void
 have_a_query (query_string, query_type, query_result, depth)
-char *query_string;
-int query_type;
-a_res *query_result;
-int depth;
+     char *query_string;
+     int query_type;
+     a_res *query_result;
+     int depth;
 {
-	int i;
-	a_res res;
-	char tmp [256];
-
-	dprintf (
+    int i;
+    a_res res;
+    char tmp [256];
+    
+    dprintf (
 	     "*** have_a_query():  string: `%s'  type: %d  server: 0x%08lx\n",
-		 query_string, query_type, (long) _res.nsaddr.sin_addr.s_addr);
-
-	query_result->type = -1;
-	query_result->n = 0;
-
-	/* for safety check here too: */
-	if (! initialized)
-	  init_stuff ();
- 
-
-	if (depth > 1)
-	{
-		dprintf ("** have_a_query(): too deep - aborted.\n");
-		return;
-	}
-
-	/*
-         * loop through every domain suffix: 
-	 */
-	for (i = -1; i < MAXDNSRCH + 1; i++)
-	{
-		if (i == -1)
-			strcpy (tmp, query_string);
-		else if (! _res.dnsrch [i])
-			break;
-		else
-			sprintf (tmp, "%s.%s", query_string, _res.dnsrch [i]);
-
-		dprintf ("** trying: `%s'...\n", tmp);
-		do_query (tmp, query_type, &res);
-		if (res.type == query_type && res.n > 0)
-		{
-			*query_result = res;
-			return;
-		}
-		/* check ptr and soa's not recursive: */
-		if (query_type == T_SOA || query_type == T_PTR)
-		{
-			*query_result = res;
-			return;
-		}
-	}
-
-	/*
-	 * seems to be unsuccessful: look for any answer:
-	 */
-	for (i = -1; i < MAXDNSRCH + 1; i++)
-	{
-		if (i == -1)
-			strcpy (tmp, query_string);
-		else if (! _res.dnsrch [i])
-			break;
-		else
-			sprintf (tmp, "%s.%s", query_string, _res.dnsrch [i]);
-
-		dprintf ("** trying: `%s'...\n", tmp);
-		do_query (tmp, query_type, &res);
-		if (res.n > 0)
-		{
-#if 1
-			/* return first answer found: */
-			*query_result = res;
-                        return;
-#else
-			query_string = tmp;
-			break;
-#endif
-		}
-	}
-	
-	if (res.n <= 0)
-	{
-		dprintf ("*** rabaeh: cannot find any answer.\n");
-		*query_result = res;
-		return;
-	}
-
-#if 1
-	dprintf ("*** rabaeh: will return anyway.\n");
+	     query_string, query_type, (long) _res.nsaddr.sin_addr.s_addr);
+    
+    query_result->type = -1;
+    query_result->n = 0;
+    
+    /* for safety check here too: */
+    if (! initialized)
+      init_stuff ();
+    
+    
+    if (depth > 1) {
+        dprintf ("** have_a_query(): too deep - aborted.\n");
 	return;
-#else
-	/*
-	 * here we could ask recursive other hosts ...
-	 * (but still wrong and not used)
-	 */
+    }
 
-	if (res.type == T_SOA) {
-		a_res tmpres;
-		do_query (res.u.str [0], T_A, &tmpres);
-		if (tmpres.type != T_A || tmpres.n <= 0)
-		{
-			dprintf ("** have_a_query(): no T_A for T_SOA\n");
-			return;
-		}
-		_res.nsaddr.sin_addr = tmpres.u.addr [0];
-		_res.nscount = 1;
-		have_a_query (query_string, query_type, query_result, 
-			      depth + 1);
-		return;
+    /*
+     * loop through every domain suffix: 
+     */
+
+    for (i = -1; i < MAXDNSRCH + 1; i++) {
+        if (i == -1)
+	  strcpy (tmp, query_string);
+	else if (! _res.dnsrch [i])
+	  break;
+	else
+	  sprintf (tmp, "%s.%s", query_string, _res.dnsrch [i]);
+	
+	dprintf ("** trying: `%s'...\n", tmp);
+	do_query (tmp, query_type, &res);
+	if (res.type == query_type && res.n > 0)
+	  {
+	    *query_result = res;
+	    return;
+	  }
+	/* check ptr and soa's not recursive: */
+	if (query_type == T_SOA || query_type == T_PTR)
+	  {
+	    *query_result = res;
+	    return;
+	  }
+      }
+    
+    /*
+     * seems to be unsuccessful: look for any answer:
+     */
+    for (i = -1; i < MAXDNSRCH + 1; i++)
+      {
+	if (i == -1)
+	  strcpy (tmp, query_string);
+	else if (! _res.dnsrch [i])
+	  break;
+	else
+	  sprintf (tmp, "%s.%s", query_string, _res.dnsrch [i]);
+	
+	dprintf ("** trying: `%s'...\n", tmp);
+	do_query (tmp, query_type, &res);
+	if (res.n > 0)
+	  {
+#if 1
+	    /* return first answer found: */
+	    *query_result = res;
+	    return;
+#else
+	    query_string = tmp;
+	    break;
+#endif
+	  }
+      }
+    
+    if (res.n <= 0)
+      {
+	dprintf ("*** rabaeh: cannot find any answer.\n");
+	*query_result = res;
+	return;
+      }
+    
+#if 1
+    dprintf ("*** rabaeh: will return anyway.\n");
+    return;
+#else
+    /*
+     * here we could ask recursive other hosts ...
+     * (but still wrong and not used)
+     */
+    
+    if (res.type == T_SOA) {
+      a_res tmpres;
+      do_query (res.u.str [0], T_A, &tmpres);
+      if (tmpres.type != T_A || tmpres.n <= 0)
+	{
+	  dprintf ("** have_a_query(): no T_A for T_SOA\n");
+	  return;
 	}
-	else 
-		dprintf ("** have_a_query(): type is %d ???\n", res.type);
+      _res.nsaddr.sin_addr = tmpres.u.addr [0];
+      _res.nscount = 1;
+      have_a_query (query_string, query_type, query_result, 
+		    depth + 1);
+      return;
+    }
+    else 
+      dprintf ("** have_a_query(): type is %d ???\n", res.type);
 #endif
 }
 
 
-
-int 
+static int 
 dns_a (hname, ip)
-char *hname;
-char *ip[];
+     char *hname;
+     char *ip[];
 {
-  a_res res;
-  int i;
+    a_res res;
+    int i;
 
-  if (isdigit (hname [0]))
-  {
-	  if (do_debug)
-		  fprintf (stderr, "dns_ptr: returning parameter...\n");
-	  ip [0] = xstrdup (hname);
-	  res.type = T_A;
-	  return res.n = 1;
-  }
-  
-  have_a_query (hname, T_A, &res, 0);
-  if (res.n < 0)
-  {
-	  ip [0] = xstrdup (res.u.str [0]);
-	  return -1;
-  }
-  
-  if (res.type != T_A)
-	  return 0;				/* XXX: error set ? */
-  
-  for (i = 0; i < res.n; i++)
-    {
-      ip [i] = xmalloc (32);
-      strcpy (ip [i], inet_ntoa (res.u.addr [i]));	/* XXX */
+    if (is_num_addr (hname)) {
+        dprintf ("* dns_a: returning parameter...\n");
+	ip [0] = xstrdup (hname);
+	res.type = T_A;
+	return res.n = 1;
     }
-  return res.n;
+  
+    have_a_query (hname, T_A, &res, 0);
+    if (res.n < 0) {
+        ip [0] = xstrdup (res.u.str [0]);
+	return -1;
+    }
+  
+    if (res.type != T_A) {
+        return 0;				/* XXX: error set ? */
+    }
+  
+    for (i = 0; i < res.n; i++) {
+        ip [i] = xmalloc (32);
+	strcpy (ip [i], inet_ntoa (res.u.addr [i]));	/* XXX */
+    }
+
+    return res.n;
 }
 
 
-int
+static int
 dns_ptr (ip, hname)
-char *ip;
-char *hname[];
+     char *ip;
+     char *hname[];
 {
-  a_res res;
-  int i, a, b, c, d;
-  char tmp [128];
+    a_res res;
+    int i, a, b, c, d;
+    char tmp [128];
 
-  if (! isdigit (ip [0]))
-  {
-	  if (do_debug)
-		  fprintf (stderr, "dns_ptr: returning parameter...\n");
-	  hname [0] = xstrdup (ip);
-	  res.type = T_PTR;
-	  return res.n = 1;
-  }
+    if (! is_num_addr (ip)) {
+        dprintf ("* dns_ptr: returning parameter (ip=`%s')...\n", ip);
+	hname [0] = xstrdup (ip);
+	res.type = T_PTR;
+	return res.n = 1;
+    }
 
-  if (4 != sscanf (ip, "%d.%d.%d.%d", &a, &b, &c, &d))
-    return 0;
+    if (is_num_addr (ip)) {
+        if (4 != sscanf (ip, "%d.%d.%d.%d", &a, &b, &c, &d)) return 0;
+	sprintf (tmp, "%d.%d.%d.%d.in-addr.arpa", d, c, b, a);
+    } else {
+        strcpy (tmp, ip);
+    }
 
-  sprintf (tmp, "%d.%d.%d.%d.in-addr.arpa", d, c, b, a);
+    have_a_query (tmp, T_PTR, &res, 0);
+    if (res.n < 0) {
+        hname [0] = xstrdup (res.u.str [0]);
+	return -1;
+    }
 
-  have_a_query (tmp, T_PTR, &res, 0);
-  if (res.n < 0)
-  {
-	  hname [0] = xstrdup (res.u.str [0]);
-	  return -1;
-  }
+    if (res.type != T_PTR) return 0;
 
-  if (res.type != T_PTR)
-	  return 0;
+    for (i = 0; i < res.n; i++) {
+        hname [i] = xstrdup (res.u.str [i]);
+    }
 
-  for (i = 0; i < res.n; i++)
-      hname [i] = xstrdup (res.u.str [i]);
-
-  return res.n;
+    return res.n;
 }
 
 
@@ -707,126 +751,159 @@ char *hname[];
 
 static void
 hinfo_clean (str)
-char *str;
+     char *str;
 {
-	char *ptr;
+     char *ptr;
 
-	while (str && *str)
-	{
-		if (*str == '\\')
-		{
-			for (ptr = str; *ptr; ptr++)
-				*ptr = *(ptr + 1);
-		}
-		str++;
-	}
+     while (str && *str) {
+         if (*str == '\\') {
+	     for (ptr = str; *ptr; ptr++)
+	       *ptr = *(ptr + 1);
+	 }
+	 str++;
+     }
 }
 
 
-int
+static int
 dns_hinfo (hname, hinfo)
-char *hname;
-char *hinfo[];
+     char *hname;
+     char *hinfo[];
 {
-  a_res res;
-  char *beg, *ptr;
+    a_res res;
+    char *beg, *ptr;
 
-  have_a_query (hname, T_HINFO, &res, 0);
-  if (res.n < 0)
-  {
-          hinfo [0] = xstrdup (res.u.str [0]);
-          return -1;
-  }
+    /*
+     * be a friend: if we get a numerical address, convert to ptr first.
+     */
 
-  if (res.type != T_HINFO)
-	  return 0;
+    if (is_num_addr (hname)) {
+        char *res;
 
- beg = ptr = res.u.str [0];
-  while (*ptr && *ptr != '.')
-    {
-      if (*ptr == '\\' && *(ptr+1))
+	if (convert_to_ptr (hname, &res) < 0) {
+  	    hinfo [0] = res;		/* error string */
+	    return -1;
+	}
+
+	hname = res;
+    }
+
+    have_a_query (hname, T_HINFO, &res, 0);
+    if (res.n < 0) {
+        hinfo [0] = xstrdup (res.u.str [0]);
+	return -1;
+    }
+
+    if (res.type != T_HINFO) return 0;
+
+    beg = ptr = res.u.str [0];
+    while (*ptr && *ptr != '.') {
+        if (*ptr == '\\' && *(ptr+1)) ptr++;
 	ptr++;
-      ptr++;
     }
   
-  if (*ptr == '.')
-	  *ptr++ = ' ';
-  while (*ptr && *ptr != '.')
-    {
-      if (*ptr == '\\' && *(ptr+1))
+    if (*ptr == '.') *ptr++ = ' ';
+    
+    while (*ptr && *ptr != '.') {
+        if (*ptr == '\\' && *(ptr+1))  ptr++;
 	ptr++;
-      ptr++;
     }
   
-  hinfo [0] = xmalloc (ptr - beg + 1);
-  strncpy (hinfo [0], beg, ptr - beg);
-  hinfo [0][ptr - beg] = 0;
-  hinfo_clean (hinfo [0]);
-
-  return 1;
+    hinfo [0] = xmalloc (ptr - beg + 1);
+    strncpy (hinfo [0], beg, ptr - beg);
+    hinfo [0][ptr - beg] = 0;
+    hinfo_clean (hinfo [0]);
+    
+    return 1;
 }
 
 
-int 
+static int 
 dns_mx (hname, mx)
-char *hname;
-char *mx[];
+     char *hname;
+     char *mx[];
 {
-	a_res res;
-	int i;
+    a_res res;
+    int i;
 
-	have_a_query (hname, T_MX, &res, 0);
-	if (res.n < 0)
-	{
-		mx [0] = xstrdup (res.u.str [0]);
-		return -1;
+    /*
+     * be a friend: if we get a numerical address, convert to ptr first.
+     */
+
+    if (is_num_addr (hname)) {
+        char *res;
+	    
+	if (convert_to_ptr (hname, &res) < 0) {
+	    mx [0] = res;		/* error string */
+	    return -1;
 	}
+	
+	hname = res;
+    }
 
-	if (res.type != T_MX)
-		return 0;
+    have_a_query (hname, T_MX, &res, 0);
+    if (res.n < 0) {
+        mx [0] = xstrdup (res.u.str [0]);
+	return -1;
+    }
 
-	for (i = 0; i < res.n; i++)
-		mx [i] = xstrdup (res.u.str [i]);
+    if (res.type != T_MX) return 0;
 
-	return res.n;
+    for (i = 0; i < res.n; i++)
+      mx [i] = xstrdup (res.u.str [i]);
+
+    return res.n;
 }
 
 
-int 
+static int 
 dns_soa (hname, soa)
-char *hname;
-char *soa[];
+     char *hname;
+     char *soa[];
 {
-	a_res res;
-	int i;
+    a_res res;
+    int i;
 
-	have_a_query (hname, T_SOA, &res, 0);
-	if (res.n < 0)
-	{
-		soa [0] = xstrdup (res.u.str [0]);
-		return -1;
+    /*
+     * be a friend: if we get a numerical address, convert to ptr first. 
+     * i guess, this makes absolut no sense for a soa :-)
+     */
+
+    if (is_num_addr (hname)) {
+        char *res;
+	    
+	if (convert_to_ptr (hname, &res) < 0) {
+	    soa [0] = res;		/* error string */
+	    return -1;
 	}
+	
+	hname = res;
+    }
 
-	if (res.type != T_SOA)
-		return 0;
+    have_a_query (hname, T_SOA, &res, 0);
+    if (res.n < 0) {
+        soa [0] = xstrdup (res.u.str [0]);
+	return -1;
+    }
 
-	for (i = 0; i < res.n; i++)
-		soa [i] = xstrdup (res.u.str [i]);
+    if (res.type != T_SOA) return 0;
 
-	return res.n;
+    for (i = 0; i < res.n; i++)
+      soa [i] = xstrdup (res.u.str [i]);
+
+    return res.n;
 }
 
 
-
 /*
  * may be this a tcl-like interface:
  */
 
-int
+static int
 dns_query (argc, argv, buf)
-int argc;
-char *argv[];
-char *buf;
+     int argc;
+     char *argv[];
+     char *buf;
 {
   int query_type = -1;			/* what to do */
   char *query_string = 0;		/* what to pass */
@@ -834,6 +911,8 @@ char *buf;
   int i, rc;
 
   dprintf ("* dns_query: called with argc: %d\n", argc);
+  { int k; for (k = 0; k < argc; k++) 
+	dprintf ("*  argv [%d] = `%s'\n", k, argv [k]); }
 
   /* default values: */
   timeout = 2;
@@ -1026,7 +1105,7 @@ char *buf;
 	    sprintf (buf + strlen (buf), "%s", result [i]);
 	    if (i + 1 < rc)
 		    strcat (buf, " ");
-	    xfree (result [i]);
+	    free (result [i]);
     }
   }      
 
@@ -1034,18 +1113,16 @@ char *buf;
 }
 
 
-#ifndef STANDALONE
-
 /* 
  * tcl called interface:
  */
 
 int
-dns_cmd (clientData, interp, argc, argv)
-ClientData clientData;
-Tcl_Interp* interp;
-int argc;
-char *argv[];
+dnsCmd (clientData, interp, argc, argv)
+     ClientData clientData;
+     Tcl_Interp *interp;
+     int argc;
+     char *argv[];
 {
 	char answer [1024];
 	int n;
@@ -1055,10 +1132,10 @@ char *argv[];
 	 */
 	if (argc < 3)
 	{
-		Tcl_AppendResult (interp, "bad # args: ", argv[0], 
-				  " ?-timeout t? ?-retries r? ?-s host? ",
-				  "a | ptr | hinfo | mx | soa ",
-				  " name-or-address", (char *) NULL);
+		Tcl_AppendResult (interp, "wrong # args: should be \"", 
+				  argv[0], 
+				  " ?-timeout t? ?-retries r? ?-server host? ",
+				  "option name-or-address \"", (char *) NULL);
 		return TCL_ERROR;
 	}
 	
@@ -1072,152 +1149,4 @@ char *argv[];
 	return (n >= 0) ? TCL_OK : TCL_ERROR;
 }
 
-#else /* STANDALONE */
-
-/*
- * a standalone main() for testing of the query-code:
- */
-
-int
-main (argc, argv)
-int argc;
-char *argv [];
-{
-	a_res res;
-	int rc, n, i;
-	static int query_what [2] = { T_A, T_HINFO };
-	char tmp [256], *query_string = 0;
-
-	if (argc > 1 && ! strcmp (argv [1], "-d"))
-	  {
-	    do_debug = 1;
-	    argc--, argv++;
-	  }
-
-	if (argc <= 2)
-	{
- 		fputs ("use: dnsfun [<opts>] <hostname>|<ipaddress>\n",
-		       stderr);
-		exit (1);
-	}
-
-	rc = dns_query (argc - 1, argv + 1, tmp);
-	printf ("got rc: %d   result: `%s'\n", rc, tmp);
-	exit (0);
-
-
-	/*
-	 * initialize resolver:
-	 */
-	if (! initialized)
-	  init_stuff ();
- 
-#if 1
-	for (i = 0; i < MAXDNSRCH + 1 && _res.dnsrch [i]; i++)
-		dprintf ("* search-order:  %d: `%s'\n", i, _res.dnsrch [i]);
-	_res.retrans = 2, _res.retry = 2;
-	dprintf ("* retrans: %d   timeout: %d\n", _res.retrans, _res.retry);
-#endif
-
-	{
-	  static char hname [256];
-
-	if (isdigit (argv [1][0]))
-	{
-	  char *res [10];
-	  int i, rc = dns_ptr (argv [1], res);
-	  printf ("got rc: %d\n", rc);
-	  if (! rc)
-	    printf ("  no info\n");
-	  for (i = 0; i < rc; i++)
-	    printf ("  `%s'\n", res [i]);
-	  strcpy (hname, res [0]);
-	}
-	else {
-	  char *res [10];
-	  int i, rc = dns_a (argv [1], res);
-	  printf ("got rc: %d\n", rc);
-	  if (! rc)
-	    printf ("  no info\n");
-	  for (i = 0; i < rc; i++)
-	    printf ("  `%s'\n", res [i]);
-	  strcpy (hname, argv [1]);
-	}
-	  
-	  if (! hname [0])
-	    exit (0);
-	  
-	  {
-	    char *res [10];
-	    int i, rc = dns_hinfo (hname, res);
-	    printf ("got rc: %d\n", rc);
-	  if (! rc)
-	    printf ("  no info\n");
-	    for (i = 0; i < rc; i++)
-	      printf ("  `%s'\n", res [i]);
-	    strcpy (hname, argv [1]);
-	  }
-	  exit (0);
-	}
-	/* XXX */
-
-	if (isdigit (argv [1][0]))
-	{
-		int a, b, c, d;
-		if (4 == sscanf (argv [1], "%d.%d.%d.%d", &a, &b, &c, &d))
-		{
-			a_res tmpres;
-			sprintf (tmp, "%d.%d.%d.%d.in-addr.arpa",
-				 d, c, b, a);
-			have_a_query (tmp, T_PTR, &tmpres, 0);
-			if (tmpres.type == T_PTR && tmpres.n > 0)
-			{
-				query_string = xstrdup (tmpres.u.str [0]);
-				printf ("got reverse lookup:\n");
-				printf ("  `%s'\n", query_string);
-			}
-		}
-	}
-
-	if (! query_string)
-		query_string = argv [1];
-
-	for (n = 0; n < 2; n++)
-	{
-		have_a_query (query_string, query_what [n], &res, 0);
-
-		printf ("query for type %d:\n", query_what [n]);
-
-		if (res.n <= 0)
-		{
-			printf ("sigh - nothing found.\n");
-		}
-		else {
-			for (i = 0; i < res.n; i++)
-			{
-				switch (res.type) {
-				case T_A:
-					printf ("  `%s' (0x%08lx)\n",
-						inet_ntoa (res.u.addr [i]),
-						res.u.addr[i].s_addr);
-					break;
-				case T_SOA:
-				case T_HINFO:
-				case T_PTR:
-					printf ("  `%s'\n", res.u.str [i]);
-					break;
-				default:
-					printf ("  ???\n");
-				}
-			}
-		}
-	}
-
-	return 0;
-}
-
-#endif /* STANDALONE */
-
-#endif /* HAVE_DNS */
-
 /* end of dnsfun.c */
