@@ -1,38 +1,41 @@
 /*
- * ntping.c:						jan 1993
+ * ntping.c:						Jan 1993
  * (schoenfr@ibr.cs.tu-bs.de)
  *
  * Copyright (c) 1993, 1994 by Erik Schoenfelder and Juergen Schoenwaelder 
- * TU Braunschweig, Germany Institute for Operating Systems 
+ * TU Braunschweig, Germany, Institute for Operating Systems 
  * and Computer Networks.
  *
- * ping/traceroute for tkined (with scotty and tcl).
+ * This is ping/traceroute for tkined (with scotty and tcl).
  */
 
 #include <stdio.h>
 #ifdef HAVE_STDLIB_H
-#include <stdlib.h>
+# include <stdlib.h>
 #endif
 #include <signal.h>
 #include <string.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#ifdef HAVE_SYS_SELECT_H
+# include <sys/select.h>
+#endif
 #include <errno.h>
 #if defined(pyr) || (defined(MACH) && defined (MTXINU))
 /* should make into the configure script... */
 extern int errno;
 #endif
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+# include <unistd.h>
 #endif
 #ifdef HAVE_MALLOC_H
-#include <malloc.h>
+# include <malloc.h>
 #endif
 #include <fcntl.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #ifdef DB_MALLOC
-#include <dbmalloc.h>
+# include <dbmalloc.h>
 #endif
 
 #ifdef __alpha
@@ -48,25 +51,34 @@ typedef unsigned long ipaddr_t;
 typedef unsigned int int32;
 
 
+/* aix failes to use IP_TTL correct: */
+#ifndef _AIX
+#if defined(IP_TTL) && ! defined(USE_DLPI)
+/*
+ * set this, if we will not send our own handmade ip (udp) packets.
+ * so we open a ordinariy udp socket and set the ttl via
+ * setsockopt with IP_TTL.
+ * This is used for HP-UX and SVR4-boxes.
+ */
+#define USE_DLPI
+#endif /* IP_TTL && ! USE_DLPI */
+#endif /* ! aix */
+
 #ifdef linux
 /*
- * this is for linux around 0.99.6 and above:
+ * this is for linux around 0.99.15 and above:
  */
 struct ip { u_char ip_hl:4, ip_v:4; u_char ip_tos; short ip_len; 
 	    u_short ip_id; short ip_off; u_char ip_ttl; u_char ip_p; 
-	    u_short ip_sum; struct  in_addr ip_src,ip_dst;
+	    u_short ip_sum; struct in_addr ip_src,ip_dst;
 };
 struct icmp { u_char  icmp_type; u_char icmp_code; u_short icmp_cksum;
 	      u_short icmp_id, icmp_seq; char icmp_data [1];
 };
 struct udphdr { u_short uh_sport, uh_dport; short uh_ulen; u_short uh_sum; };
-/*
- * linux with new net-2 networking code: 
- */
+
 #include <linux/wait.h>
-/** #include <linux/udp.h> **/
-/** #include <linux/tcp.h> **/
-/** #include <linux/icmp.h> **/
+
 #define ICMP_MINLEN     	8
 #define ICMP_ECHO       	8
 #define ICMP_ECHOREPLY       	0
@@ -92,7 +104,7 @@ struct udphdr { u_short uh_sport, uh_dport; short uh_ulen; u_short uh_sum; };
 
 #define PROGNAME	"ntping"		/* my name */
 
-static char *version = "ntping;  v0.6b  Dec 1993";
+static char *version = "ntping  v0.8f  November 1994";
 
 /* length of the data space of a ping packet: */
 #define DEFAULT_DATALEN		44
@@ -135,6 +147,13 @@ static int data_len = DEFAULT_DATALEN;
 /* base port for ttl probes: */
 static int base_port = BASE_PORT;
 
+#ifdef USE_DLPI
+/* port # our udp-socket is bound to: */
+static unsigned short src_port;
+/* second ip-socket: */
+static int ipsock2 = -1;
+#endif
+
 /* read hostnames from stdin: */
 static int interactive = 0;
 
@@ -156,9 +175,11 @@ static int reply_same = 0;
 /* milli-seconds to sleep if we have sent a packet: */
 static int delay_time = 0;			/* default: don't sleep */
 
-/* *sigh* check for bytesex in unrech-icmp's too. the synoptics are dumb
-   enough to reply wrong packets */
+/* sigh - check for bytesex in unreach/exceed icmp's too. 
+   the synoptics are dumb enough to reply wrong packets: */
 static int check_sex = 1;
+/* emit a warning, if a byte-swapped port in a reply is found: */
+static int warn_swapped = 1;
 
 /* bones argument string: */
 static char bbuf [20480];
@@ -192,13 +213,19 @@ static a_job *jobs = 0;
  * udp ttl packet:
  */
 typedef struct _a_pkt {
-	struct ip ip;
-	struct udphdr udph;
-	/* arg: these fields are not replied ... should be removed. */
-	char seq;
-	char ttl;
-	struct timeval tv;
-	char buf [1];
+#ifndef USE_DLPI
+    /* 
+     * we are sending simple udp packets; no additional headers are
+     * needed.
+     */
+    struct ip ip;
+    struct udphdr udph;
+#endif
+    /* arg: no additional bytes are replied icmp - this is nonsense. */
+    char seq;
+    char ttl;
+    struct timeval tv;
+    char buf [1];
 } a_pkt;
 
 
@@ -264,7 +291,7 @@ unsigned n;
 
 	if (select (1, 0, 0, 0, &tv) < 0)
 		if (do_debug)
-                        perror ("m_usleep: select failed; reason");
+                        perror ("ntping: m_usleep: select failed; reason");
 }
 
 #endif /* NEED_USLEEP */
@@ -278,6 +305,8 @@ unsigned n;
  * how to have a short delay (1 to 20 ms ?):
  * 1. loop with gettimeofday's() and compare the wasted time.
  * 2. calc the delay of a empty loop and use this to waste time.
+ *
+ * both is ugly, but what should i do ?
  */
 
 /*
@@ -410,8 +439,9 @@ struct sockaddr_in *addr;
 {
 	static char buf [99];
 	ipaddr_t x = ntohl (addr->sin_addr.s_addr);
-	sprintf (buf, "%d.%d.%d.%d", (x >> 24) & 0xff, (x >> 16) & 0xff,
-		 (x >> 8) & 0xff, x & 0xff);
+	sprintf (buf, "%d.%d.%d.%d", (int) (x >> 24) & 0xff, 
+		 (int) (x >> 16) & 0xff,
+		 (int) (x >> 8) & 0xff, (int) x & 0xff);
 	return buf;
 }
 
@@ -456,34 +486,73 @@ unsigned x;
 }
 
 
+/*
+ * search for a specific job.
+ * the id is either the src-port of our own udp-packet (== job->id) or
+ * for the dlpi udp-packet the static src_port.
+ */
+
 static a_job *
-find_job_port (id, port)
-unsigned short id;
-unsigned short port;
+find_job_port (ip, udph)
+struct ip *ip;
+struct udphdr *udph;
 {
-	a_job *job;
+    unsigned short id = ntohs (udph->uh_sport);
+    unsigned short port = ntohs (udph->uh_dport);
+    a_job *job;
+    
+    if (do_debug)
+      fprintf (stderr, "* looking for src %u (0x%lx)  dest %u (0x%lx) ...", 
+	       (unsigned) id, (int) id, (unsigned) port, (int) port);
 
-	if (do_debug)
-		fprintf (stderr, "* looking for job id %u port %u ...", 
-			 (unsigned) id, (unsigned) port);
+    for (job = jobs; job; job = job->next)
+      {
+	  unsigned short src;
+	  int got_it = 0;
 
-	for (job = jobs; job; job = job->next)
-	{
+#ifndef USE_DLPI
+	  src = job->id;
+#else
+	  src = src_port;
+#endif
+
+	  if (do_debug)
+#ifndef USE_DLPI
+	    fprintf (stderr, " %u", src);
+#else
+	    fprintf (stderr, "  %u (0x%lx)", (unsigned) job->port, 
+		     (int) job->port);
+#endif
+	  
+	  if (job->port == port && src == id)
+	    got_it = 1;
+	  
+	  if (! got_it && check_sex)
+	    {
+		if ((job->port == sexy (port) && src == sexy (id))
+		    || (job->port == port && src == sexy (id))
+		    || (job->port == sexy (port) && src == id))
+		  {
+		      got_it = 1;
+		      if (warn_swapped)
+			fprintf (stderr,
+"ntping: warning: got icmp-reply from 0x%08lx with byte-swapped port in the reply\n", 
+				 ntohl (ip->ip_src.s_addr));
+		  }
+	    }
+	  
+	  if (got_it)
+	    {
 		if (do_debug)
-			fprintf (stderr, " %u", (unsigned) job->id);
-		if ((job->port == port && job->id == id)
-		    || (check_sex && job->port == sexy (port)
-			&& job->id == sexy (id)))
-		{
-			if (do_debug)
-				fprintf (stderr, " got it.\n");
-			return job;
-		}
-	}
+		  fprintf (stderr, " got it.\n");
+		return job;
+	    }
+      }
 
-	if (do_debug)
-		fprintf (stderr, " nope...\n");
-	return (a_job *) 0;
+    if (do_debug)
+      fprintf (stderr, " nope...\n");
+
+    return (a_job *) 0;
 }
 
 
@@ -524,7 +593,7 @@ make_addr (addr, hname)
 struct sockaddr_in *addr;
 char *hname;
 {
-	bzero ((char *) addr, sizeof (struct sockaddr_in));
+	memset ((char *) addr, 0, sizeof (struct sockaddr_in));
 	if (*hname >= '0' && *hname  <= '9')
 	{
 		int a, b, c, d;
@@ -545,7 +614,7 @@ char *hname;
 		if (hp)
 		{
 			addr->sin_family = hp->h_addrtype;
-			bcopy ((char *) hp->h_addr, (char *) &(addr->sin_addr),
+			memcpy ((char *) &(addr->sin_addr), (char *) hp->h_addr,
 			       hp->h_length);
 			return 1;
 		}
@@ -567,223 +636,227 @@ static void
 do_receive (stime)
 int stime;
 {
-	char packet [MAX_DATALEN + 128];
-        int len = sizeof (packet);
-	struct ip *ip = (struct ip *) packet;
-	struct icmp *icp;		/* for pings */
-	struct udphdr *udph;		/* for ttl's */
-        struct sockaddr_in sfrom;
-	struct timeval tp1, tp2;
-        int fromlen = sizeof (sfrom);
-	int hlen = 0, is_ttl = 0, is_mask, is_tstamp, cc, rc, ttl_is_done;
-	int do_continue;
-	ipaddr_t val = 0;
-	a_job *job = 0;
-	
-	for (;;) 
-	{
-		rc = fd_ready (icsock, stime);		
-		if (! rc)
-			return;
-		
-		cc = recvfrom (icsock, (char *) packet, len, 0,
-			       (struct sockaddr *) &sfrom, &fromlen);
-
-		if (gettimeofday (&tp2, (struct timezone *) 0) == -1) 
-		{
-			perror ("ntping: gettimeofday");
-			continue;
-		}
-		
+    char packet [MAX_DATALEN + 128];
+    int len = sizeof (packet);
+    struct ip *ip = (struct ip *) packet;
+    struct icmp *icp;		/* for pings */
+    struct udphdr *udph;		/* for ttl's */
+    struct sockaddr_in sfrom;
+    struct timeval tp1, tp2;
+    int fromlen = sizeof (sfrom);
+    int hlen = 0, is_ttl = 0, is_mask, is_tstamp, cc, rc, ttl_is_done;
+    int do_continue;
+    ipaddr_t val = 0;
+    a_job *job = 0;
+    
+    for (;;) 
+      {
+	  rc = fd_ready (icsock, stime);		
+	  if (! rc)
+	    return;
+	  
+	  cc = recvfrom (icsock, (char *) packet, len, 0,
+			 (struct sockaddr *) &sfrom, &fromlen);
+	  
+	  if (gettimeofday (&tp2, (struct timezone *) 0) == -1) 
+	    {
 		if (do_debug)
-			printf ("... recv got: rc %d\n", cc);
-
-		if (cc < 0 && errno != EINTR)
-		{
-			perror ("ping: recvfrom");
-			continue;
-		}
-		
-		if (cc == 0)
-			return;
-
-		/* how to play this game correct ???
-		 * raw-socket with icmp-protocol should send and receive
-		 * ip packets with complete header:
-		 */
-#ifdef linux
-		/* still not correct in net-2e-alpha-2: ... */
-		/* and in net-2e-alpha-3 too :-( */
+		  perror ("ntping: gettimeofday");
+		return;
+	    }
+	  
+	  if (do_debug)
+	    printf ("... recv got: rc %d\n", cc);
+	  
+	  if (cc < 0 && errno != EINTR)
+	    {
+		if (do_debug)
+		  perror ("ntping: warning: recvfrom");
+		return;
+	    }
+	  
+	  if (cc == 0)
+	    return;
+	  
+	  /* how to play this game correct ???
+	   * raw-socket with icmp-protocol should send and receive
+	   * ip packets with complete header:
+	   */
+#ifdef old_and_broken_linux
+	  /* still not correct in net-2e-alpha-2: ... */
+	  /* and in net-2e-alpha-3 too :-( */
 #define IP_VERS		4
 #define IP_PKTLEN	(sizeof (struct ip) / sizeof (int))
-
-		if (ip->ip_v != IP_VERS && ip->ip_hl != IP_PKTLEN)
-		{
-			/* seems to be a missing ip-header: */
-			if (do_debug)
-				fprintf (stderr, 
-					 "* linux: missing ip-header ...\n");
-		}
-		else
-#else
-		hlen = ip->ip_hl << 2;
-#endif
-		if (cc < hlen + ICMP_MINLEN)
-			continue;
-
-		icp = (struct icmp *) (packet + hlen);
-		
-		udph = (struct udphdr *) (icp->icmp_data + sizeof (struct ip));
-
+	  
+	  if (ip->ip_v != IP_VERS && ip->ip_hl != IP_PKTLEN)
+	    {
+		/* seems to be a missing ip-header: */
 		if (do_debug)
-			fprintf (stderr,
-			 "... got icmp type %d  code %d  id %u (0x%x)\n", 
-				 icp->icmp_type, icp->icmp_code,
-				 (unsigned) icp->icmp_id,
-				 (unsigned) icp->icmp_id);
-
-		do_continue = is_tstamp = is_mask = ttl_is_done = 0;
-
-		switch (icp->icmp_type) {
-		case ICMP_ECHOREPLY:
-			if (do_debug)
-				fprintf (stderr, "* got ECHOREPLY.\n");
-			is_ttl = 0;
-			break;
-		case ICMP_MASKREPLY:
-			if (do_debug)
-                                fprintf (stderr, "* got MASKREPLY.\n");
-			is_ttl = 0, is_mask = 1;
-			break;
-		case ICMP_TSTAMPREPLY:
-			if (do_debug)
-                                fprintf (stderr, "* got TSTAMPREPLY.\n");
-			is_ttl = is_mask = 0; is_tstamp = 1;
-			break;
-		case ICMP_UNREACH:
-			if (do_debug)
-				fprintf (stderr, "* got UNREACHABLE.\n");
-			if (icp->icmp_code != ICMP_UNREACH_PORT)
-			{
-				if (do_debug)
-					fprintf (stderr,
-						 "* bad code - discarded.\n");
-				do_continue = 1;
-			}
-			is_ttl = 1;
-			ttl_is_done = 1;
-			break;
-		case ICMP_TIMXCEED:
-			if (do_debug)
-				fprintf (stderr, "* got TIMXCEED_INTRANS.\n");
-			if (icp->icmp_code != ICMP_TIMXCEED_INTRANS)
-			{
-				if (do_debug)
-					fprintf (stderr,
-						 "* bad code - discarded.\n");
-				do_continue = 1;
-			}
-			is_ttl = 1;
-			break;
-		case ICMP_SOURCEQUENCH:
-			{ struct ip *ip = (struct ip *) ((char *) icp + 4);
-			  if (do_debug)
-				fprintf (stderr, 
-			"* got icmp sourcequench from 0x%lx for 0x%lx\n",
-					 (long) ip->ip_dst.s_addr,
-					 (long) ip->ip_src.s_addr);
- 			}
-			/* fall through */
-		default:
-			if (do_debug)
-			  fprintf (stderr, "* packet discarded.\n");
-			do_continue = 1;
-		}
-
-		if (do_continue == 1)
-			continue;
-
-		/* XXX: oops bert nase: why this ? */
-		if (rc <= 0)
-			rc = 1;
-		
-		if ((! is_ttl && ! (job = find_job_id (icp->icmp_id)))
-		    || (is_ttl && ! (job = find_job_port (udph->uh_sport,
-							  udph->uh_dport))))
-			continue;
-
-		/* this one is still done: */
-		if (job->done)
+		  fprintf (stderr, 
+			   "* linux: missing ip-header ...\n");
+	    }
+	  else
+	    /* in linux 0.99.15a this is fixed - hurray ! */
+#else
+	    hlen = ip->ip_hl << 2;
+#endif
+	  if (cc < hlen + ICMP_MINLEN)
+	    continue;
+	  
+	  icp = (struct icmp *) (packet + hlen);
+	  
+	  udph = (struct udphdr *) (icp->icmp_data + sizeof (struct ip));
+	  
+	  if (do_debug)
+	    fprintf (stderr,
+		     "... got icmp type %d  code %d  id %u (0x%x)\n", 
+		     icp->icmp_type, icp->icmp_code,
+		     (unsigned) icp->icmp_id,
+		     (unsigned) icp->icmp_id);
+	  
+	  do_continue = is_tstamp = is_mask = ttl_is_done = 0;
+	  
+	  switch (icp->icmp_type) {
+	    case ICMP_ECHOREPLY:
+	      if (do_debug)
+		fprintf (stderr, "* got ECHOREPLY.\n");
+	      is_ttl = 0;
+	      break;
+	    case ICMP_MASKREPLY:
+	      if (do_debug)
+		fprintf (stderr, "* got MASKREPLY.\n");
+	      is_ttl = 0, is_mask = 1;
+	      break;
+	    case ICMP_TSTAMPREPLY:
+	      if (do_debug)
+		fprintf (stderr, "* got TSTAMPREPLY.\n");
+	      is_ttl = is_mask = 0; is_tstamp = 1;
+	      break;
+	    case ICMP_UNREACH:
+	      if (do_debug)
+		fprintf (stderr, "* got UNREACHABLE.\n");
+	      if (icp->icmp_code != ICMP_UNREACH_PORT)
 		{
-			if (do_debug)
-				fprintf (stderr, "* still done :-)\n");
-			continue;
+		    if (do_debug)
+		      fprintf (stderr,
+			       "* bad code - discarded.\n");
+		    do_continue = 1;
 		}
-
-		if (is_ttl != (job->tag == 't'))
-			fprintf (stderr, "oops - ping or ttl ? (tell nase)\n");
-
-		if (! is_ttl)
+	      is_ttl = 1;
+	      ttl_is_done = 1;
+	      break;
+	    case ICMP_TIMXCEED:
+	      if (do_debug)
+		fprintf (stderr, "* got TIMXCEED_INTRANS.\n");
+	      if (icp->icmp_code != ICMP_TIMXCEED_INTRANS)
 		{
-			if (is_mask)
-			{
-				val = ntohl (* (ipaddr_t *) icp->icmp_data);
-				job->mask = val;
-				tp1 = job->tv;
-			}
-			else if (is_tstamp)
-			  {
-			    if (do_debug) {
-				    fprintf (stderr,
-					     "** tstamp: got %d, %d, %d\n",
+		    if (do_debug)
+		      fprintf (stderr,
+			       "* bad code - discarded.\n");
+		    do_continue = 1;
+		}
+	      is_ttl = 1;
+	      break;
+	    case ICMP_SOURCEQUENCH:
+	      { struct ip *ip = (struct ip *) ((char *) icp + 4);
+		if (do_debug)
+		  fprintf (stderr, 
+			   "* got icmp sourcequench from 0x%lx for 0x%lx\n",
+			   (long) ip->ip_dst.s_addr,
+			   (long) ip->ip_src.s_addr);
+	    }
+	      /* fall through */
+	    default:
+	      if (do_debug)
+		fprintf (stderr, "* packet discarded.\n");
+	      do_continue = 1;
+	  }
+	  
+	  if (do_continue == 1)
+	    continue;
+	  
+	  /* XXX: oops bert nase: why this ? */
+	  if (rc <= 0)
+	    rc = 1;
+	  
+	  if ((! is_ttl && ! (job = find_job_id (icp->icmp_id)))
+	      || (is_ttl && ! (job = find_job_port (ip, udph))))
+	    continue;
+	  
+	  /* this one is still done: */
+	  if (job->done)
+	    {
+		if (do_debug)
+		  fprintf (stderr, "* still done :-)\n");
+		continue;
+	    }
+	  
+	  if (is_ttl != (job->tag == 't'))
+	    fprintf (stderr, "oops - ping or ttl ? (tell nase)\n");
+	  
+	  if (! is_ttl)
+	    {
+		if (is_mask)
+		  {
+		      val = ntohl (* (ipaddr_t *) icp->icmp_data);
+		      job->mask = val;
+		      tp1 = job->tv;
+		  }
+		else if (is_tstamp)
+		  {
+		      if (do_debug) {
+			  fprintf (stderr,
+				   "** tstamp: got %d, %d, %d\n",
 			           ntohl (((int32 *) icp->icmp_data) [0]),
 			           ntohl (((int32 *) icp->icmp_data) [1]),
 			           ntohl (((int32 *) icp->icmp_data) [2]));
-			      fprintf (stderr,
-				       "** tstamp: diff is %d\n",
+			  fprintf (stderr,
+				   "** tstamp: diff is %d\n",
 			           ntohl (((int32 *) icp->icmp_data) [1]) -
 			           ntohl (((int32 *) icp->icmp_data) [0]));
-		            }
-			    
-			    val = ntohl (((int32 *) icp->icmp_data) [1]) -
-			      ntohl (((int32 *) icp->icmp_data) [0]);
-
-			    job->mask = val;
-			    tp1 = job->tv;
-			  }
-			else
-				bcopy ((char *) icp->icmp_data, (char *) &tp1, 
-				       sizeof (struct timeval)); 
-		}
-		else {
-			tp1 = job->tv;
-			job->tdone = ttl_is_done;
-		}
-
-		rc = (tp2.tv_sec - tp1.tv_sec) * 1000 
-			+ (tp2.tv_usec - tp1.tv_usec) / 1000;
-
-		job->tim = (rc <= 0) ? 1 : rc;
-		job->done = 1;
-		job->hop = sfrom;
-
-		if (do_debug)
-		{
-			fprintf (stderr, "* .. got host %s (with tim %d)\n", 
-				 job->hname, job->tim);
-			if (is_mask)
-				printf ("* .. and mask 0x%lx\n", (long) val);
-			if (is_tstamp)
-			  	printf ("* .. and tstamp %ld\n", (long) val);
-			if (job->tag == 't')
-			{
-				ipaddr_t l = sfrom.sin_addr.s_addr;
-				fprintf (stderr, "* ttl hop is %s (0x%lx)\n", 
-					 my_inaddr (&sfrom), (long) l);
-			}
-		}
-	}
-	/* not reached */
+		      }
+		      
+		      val = ntohl (((int32 *) icp->icmp_data) [1]) -
+			ntohl (((int32 *) icp->icmp_data) [0]);
+		      
+		      job->mask = val;
+		      tp1 = job->tv;
+		  }
+		else
+		  memcpy ((char *) &tp1, (char *) icp->icmp_data,
+			  sizeof (struct timeval)); 
+	    }
+	  else {
+	      tp1 = job->tv;
+	      job->tdone = ttl_is_done;
+	  }
+	  
+	  rc = (tp2.tv_sec - tp1.tv_sec) * 1000 
+	    + (tp2.tv_usec - tp1.tv_usec) / 1000;
+	  
+	  job->tim = (rc <= 0) ? 1 : rc;
+	  job->done = 1;
+	  job->hop = sfrom;
+	  
+	  if (do_debug)
+	    {
+		fprintf (stderr, "* .. got host %s (with tim %d)\n", 
+			 job->hname, job->tim);
+		if (is_mask)
+		  fprintf (stderr, "* .. and mask 0x%lx\n",
+			   (long) val);
+		if (is_tstamp)
+		  fprintf (stderr, "* .. and tstamp %ld\n",
+			   (long) val);
+		if (job->tag == 't')
+		  {
+		      ipaddr_t l = sfrom.sin_addr.s_addr;
+		      fprintf (stderr, "* ttl hop is %s (0x%lx)\n", 
+			       my_inaddr (&sfrom), (long) l);
+		  }
+	    }
+      }
+    /* not reached */
 }
 
 
@@ -824,19 +897,24 @@ a_job *job;
 {
 	char *datap, outpack [2048];
 	a_pkt *pkt = (a_pkt *) outpack;
+#ifndef USE_DLPI
 	struct ip *ip = &pkt->ip;
 	struct udphdr *udph = &pkt->udph;
+#endif /* ! USE_DLPI */
 	struct sockaddr_in *sto = &job->sa;
 	int i, j;
 	
 	if (job->done)
 	    return;
 
+#ifndef USE_DLPI
 	ip->ip_hl = sizeof (struct ip) / sizeof (int32);
 	ip->ip_v = 4;			/* take this - live and in color */
 	ip->ip_tos = 0;
-	ip->ip_id = htons (job->id);	/* ??? */
+	ip->ip_id = job->id;		/* ??? */
 	ip->ip_sum = 0;
+	/* fixed by Jan L. Peterson (jlp@math.byu.edu): */
+        ip->ip_src.s_addr = 0;          /* jlp */
 
 	ip->ip_off = 0;
 	ip->ip_p = IPPROTO_UDP;
@@ -854,12 +932,17 @@ a_job *job;
 	udph->uh_ulen = htons ((u_short) (data_len - sizeof (struct ip)));
 	udph->uh_sum = 0;
 
+#else /* USE_DLPI */
+	sto->sin_port = htons (job->port);
+#endif /* USE_DLPI */
+
 	pkt->seq = job->cur_seq++;
 	pkt->ttl = job->ttl;	
 	if (gettimeofday (&pkt->tv, (struct timezone *) 0) == -1) 
 	{
+	    if (do_debug)
 		perror ("ntping: gettimeofday");
-		return;
+	    return;
 	}
 	job->tv = pkt->tv;
 
@@ -867,18 +950,38 @@ a_job *job;
 	for (i = sizeof (struct timeval) + 2, j = 'A'; i < data_len; i++, j++)
 		datap [i] = j;
 
+/*
+ * may be simply: #ifdef IP_TTL  (eventually wrapped around a USE_DLPI) ?
+ */
+#ifdef USE_DLPI
+	{ int opt_ttl = job->ttl;
+	  int opt_ttl_len = sizeof (opt_ttl);
+
+	  if (setsockopt(ipsock, IPPROTO_IP, IP_TTL, 
+			 &opt_ttl, opt_ttl_len) < 0)
+	      perror ("ntping: setsockopt: cannot set ttl; reason");
+
+	  if (getsockopt (ipsock, IPPROTO_IP, IP_TTL,
+			  &opt_ttl, &opt_ttl_len) < 0)
+	      perror ("ntping: getsockopt: ttl not set; reason");
+	  else if (job->ttl != opt_ttl)
+	      fprintf (stderr, "ntping: dlpi: cannot set ttl - ouch...\n");
+        }
+#endif
+
 	/* fprintf(stderr, "data_len %d\n", data_len); */	/* karl */
 	i = sendto (ipsock, (char *) outpack, data_len, 0, 
 		    (struct sockaddr *) sto, sizeof (struct sockaddr_in));
 	
 	if (i < 0 || i != (data_len))  
-	  if (do_debug)
-	    perror ("ntping: sendto");
-
-	if (do_debug)
-		fprintf (stderr, "* ttl %d sent to %s  port %u (0x%x)...\n",
-			 job->ttl, job->hname, (unsigned) job->port,
-			 (unsigned) job->port);
+	  {
+	    if (do_debug)
+	      perror ("ntping: sendto");
+	  }
+	else if (do_debug)
+	  fprintf (stderr, "* ttl %d sent to %s  port %u (0x%x)...\n",
+		   job->ttl, job->hname, (unsigned) job->port,
+		   (unsigned) job->port);
 }
 
 
@@ -910,14 +1013,14 @@ a_job *job;
 	icp->icmp_id = job->id;
 
 	if (! do_mask)
-	{
-		if (gettimeofday (&tv, (struct timezone *) 0) == -1)
-		{
-			perror ("ntping: gettimeofday");
-			return;
-		}
-	}
-
+	  {
+	    if (gettimeofday (&tv, (struct timezone *) 0) == -1)
+	      {
+		if (do_debug)perror ("ntping: gettimeofday");
+		return;
+	      }
+	  }
+	
 	if (do_tstamp)
 	  {
 	    * (int32 *) datap = htonl ((tv.tv_sec % 86400) * 1000
@@ -942,15 +1045,16 @@ a_job *job;
 		    (struct sockaddr *) sto, sizeof (struct sockaddr_in));
 	
 	if (i < 0 || i != data_len)  
-	  if (do_debug)
-	    perror ("ntping: sendto");
-
-	if (do_debug)
-		fprintf (stderr, "* %s ping %d sent to %s...\n",
-			 (do_mask) ? "mask" : 
-			 ((do_tstamp) ? "tstamp" : "regular"),
-			 job->cur_seq - 1, job->hname);
-
+	  {
+	    if (do_debug)
+	      perror ("ntping: sendto");
+	  }
+	else if (do_debug)
+	  fprintf (stderr, "* %s ping %d sent to %s...\n",
+		   (do_mask) ? "mask" : 
+		   ((do_tstamp) ? "tstamp" : "regular"),
+		   job->cur_seq - 1, job->hname);
+	
 	/* lets wait a little time: */
 	if (delay_time)
 	  have_delay (delay_time);
@@ -1207,6 +1311,12 @@ char *buf;
 		 }
 		else if (! strncmp (buf, "-same", strlen ("-same")))
 			 reply_same = 1;
+		else if (! strncmp (buf, "-nocheckswap", 
+				    strlen ("-nocheckswap")))
+			check_sex = 0;
+		else if (! strncmp (buf, "-checkswap", 
+				    strlen ("-checkswap")))
+			check_sex = 1;
 		else {
 			fprintf (stderr, "ntping: bad option in `%s'\n", buf);
 			break;
@@ -1258,9 +1368,9 @@ static int
 got_bones_result ()
 {
 	a_job *j, **job = &jobs;
-	char *res = xmalloc (1024);
-	int res_len = 1024, rlen = 0, i;
-	
+	char *res;
+	int res_len, rlen, i;
+
 	do_receive (1);		/* check queue */
 	
 	for (j = jobs; j; j = j->next)
@@ -1278,6 +1388,11 @@ got_bones_result ()
 		}
 	}
 	
+	rlen = 0;
+	res_len = 1024;
+	res = xmalloc (1024);
+	*res = 0;		/* init to EMPTY result */
+
 	for (i = 0; jobs; i++)
 	{
 		for (job = &jobs; *job;  )
@@ -1535,9 +1650,29 @@ init_socks ()
 #ifdef IP_HDRINCL
 	int on = 1; /* karl */
 #endif
+#endif
+#ifdef USE_DLPI
+	struct sockaddr_in tmp_addr;
+	int ta_len = sizeof (tmp_addr);
 #endif
 	int init_data_len = data_len + sizeof (struct ip);
 
+
+	/*
+	 * sanity: dont tell about cannot open socket, if no root
+	 * permissions avail. check uid by hand.
+	 */
+
+	if (geteuid ())
+	  {
+	      fprintf (stderr,
+ "ntping: warning: running with euid %d -- not with root permissions.\n",
+		       geteuid ());
+	      fprintf (stderr,
+ "ntping: warning: expect missing permissions getting the icmp socket.\n");
+	  }
+
+
 	if ((proto = getprotobyname ("icmp")) == NULL) 
 		{ /* perror ("ntping: icmp protocol unknown; reason"); */ }
 	else
@@ -1548,17 +1683,28 @@ init_socks ()
 		perror ("ntping: cannot get icmp socket; reason");
 		return 0;
 	}
+#ifdef USE_DLPI
+	/*
+	 * My mom told me: if it hurts don't do it ...
+	 * this looks quite useable.
+	 */
+	if ((ipsock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) 
+	{
+		perror ("ntping: cannot get udp socket; reason");
+		return 0;
+	}
+#else /* ! USE_DLPI */
 	if ((ipsock = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) 
 	{
 		perror ("ntping: cannot get ip socket; reason");
 		return 0;
 	}
-
+#endif /* ! USE_DLPI */
 	/*
 	 * SO_SNDBUF and IP_HDRINCL fix from karl@sugar.NeoSoft.COM:
 	 * seems to be neccesary for 386bsd and others.
 	 */
-#ifdef SO_SNDBUF
+#if defined(SO_SNDBUF)
 	/* in interactive mode, we can have varying sizes: */
 	if (interactive)
 		init_data_len = MAX_DATALEN + sizeof (struct ip);
@@ -1569,7 +1715,7 @@ init_socks ()
 	}
 #endif /* SO_SNDBUF */
 
-#ifndef linux
+#if !defined(linux) && !defined(USE_DLPI)
 #ifdef IP_HDRINCL
 	if (setsockopt(ipsock, IPPROTO_IP, IP_HDRINCL, (char *)&on,
 		       sizeof(on)) < 0) {
@@ -1577,13 +1723,13 @@ init_socks ()
 		exit(6);
 }
 #endif /* IP_HDRINCL */
-#endif /* ! linux */
+#endif /* ! linux && ! USE_DLPI */
 
-	bzero ((char *) &maddr, sizeof (maddr));
+	memset ((char *) &maddr, 0, sizeof (maddr));
 	maddr.sin_family = AF_INET;
 	maddr.sin_addr.s_addr = INADDR_ANY;
 	maddr.sin_port = 0;
-#ifdef linux
+#if defined(linux) && ! defined(USE_DLPI)
 	/* linux pukes on bind, but seems to do the tracing stuff.
 	 * try it on your own risk.... */
 #else
@@ -1593,6 +1739,61 @@ init_socks ()
 	    return 0;
 	  }
 #endif
+
+#ifdef USE_DLPI
+	/*
+	 * we cannot send a datagramm with a src-port of our own 
+	 * choice - too bad.
+	 * therefore we connot identfify the icmp-port-unreachable 
+	 * with our homebrewed id. save the original port and check about:
+	 */
+	
+	if (getsockname (ipsock, (struct sockaddr *) &tmp_addr, &ta_len) < 0)
+	{
+	    perror ("ntping: cannot get sockname; reason");
+	    return 0;
+	}
+	src_port = ntohs (tmp_addr.sin_port);
+	if (do_debug)
+	    printf ("* dlpi: my port is %d (0x%lx)\n", (int) src_port,
+		    (int) src_port);
+
+	/*
+	 * let's try to work around a funny bug: open the complement
+	 * port, to allow reception of icmp-answers with byte-swapped
+	 * src-port field.  if we cannot get this port, silently
+	 * ignore and continue normal.
+	 */
+
+	if ((ipsock2 = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) 
+	{
+	    if (do_debug)
+	      perror ("* cannot get udp socket #2; reason");
+	    ipsock2 = -1;
+	}
+	else {
+	    struct sockaddr_in maddr2;
+
+	    memset ((char *) &maddr2, 0, sizeof (maddr2));
+	    maddr2.sin_family = AF_INET;
+	    maddr2.sin_addr.s_addr = INADDR_ANY;
+	    maddr2.sin_port = htons (sexy (src_port));
+	    
+	    if (bind (ipsock2, (struct sockaddr *) &maddr2, 
+		      sizeof (maddr2)) < 0)
+	      {
+		  if (do_debug)
+		    perror ("* cannot bind socket #2; reason");
+		  ipsock2 = -1;
+	      }
+	}
+
+	if (ipsock2 != -1 && do_debug)
+	  printf ("* got the second port # %d (0x%lx)\n", sexy (src_port),
+		  sexy (src_port));
+
+#endif /* USE_DLPI */
+
 	/* everything is fine: */
 	return 1;
 }
@@ -1721,6 +1922,10 @@ char *argv[];
 			do_tstamp = 1;
 		else if (! strcmp ("-same", *argv))
 			reply_same = 1;
+		else if (! strcmp ("-nocheckswap", *argv))
+			check_sex = 0;
+		else if (! strcmp ("-checkswap", *argv))
+			check_sex = 1;
 		else if (*argv[0] == '-')
 			usage ();
 		else if (bones)
